QuestGather = LibStub("AceAddon-3.0"):NewAddon("QuestGather", "AceConsole-3.0", "AceEvent-3.0");
local core = QuestGather;

--[[local options = {
	name = "QuestGather",
	type = 'group',
	args = {
	},
};]]
local defaults = {
	profile = {
		debug = 0,
	},
	global = {
		quests = {},
		questsNoID = {}, -- for storing quests by name before the ID is apparent
		questWarnThresh = 50,
		--debug = 0,
		zoneMap = {},
	},
};
local FACTION_MAP = {
	Alliance = 1,
	Horde = 2,
};
local output;

core.questCt = 0; -- quests logged since last reload

function core:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("QuestGatherDB", defaults, "Default");
	self.questWarnThresh = self.db.global.questWarnThresh;
	-- self.db.global.zoneMap = output;
	--[[
		LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("", options);
		local ACD = LibStub("AceConfigDialog-3.0");
		ACD:AddToBlizOptions("", "");
		self:SetEnabledState(self.db.profile.enabled;
	]]
	
	-- check if QuestHubber needs our data
	-- local qh = QuestHubber;
	-- if (qh and not qh.data or not qh.data.quests or not next(qh.data.quests)) then
		-- qh.data.quests = self.db.global.quests;
		-- qh:InitializeLoader();
	-- end
end

function core:OnEnable()
	self:RegisterEvent("GOSSIP_SHOW"); 		-- quest listings from NPCs
	self:RegisterEvent("QUEST_DETAIL");		-- single quest information
	self:RegisterEvent("QUEST_COMPLETE"); 	-- quest completion
	
	self:RegisterEvent("LOOT_OPENED"); 		-- quests from drops
	self:RegisterEvent("QUEST_ACCEPTED"); 	-- check db for fillins
end

function core:OnDisable()
	
end

function core:SlashHandle(message)

end

-- --------------------------------------------
-- Data Gathering
-- --------------------------------------------

function core:GOSSIP_SHOW()
	local npcInfo = self:UnitInfo("npc") or self:UnitInfo("questnpc");
	
	-- list of quests to pick from, cache all
	local i = 1;
	repeat
		local name, level, isTrivial, isDaily, isRepeatable = select(i, GetGossipAvailableQuests());
		
		if ( name ) then
			self:AddQuest({
				name = name,
				level = level,
				daily = isDaily,
				begin = npcInfo,
				type = "giver",
			});
		end
		
		i = i + 5;
	until (not name);
	
	-- list of quests to turn in, add data
	local i = 1;
	repeat
		local name, level, isTrivial, isDaily = select(i, GetGossipActiveQuests());
		
		if (name) then
			local id = self:GetQuestID(name);
			
			if (not id) then
				-- self:Print(format("Cannot find quest id of '%s' from log. You're probably clicking too fast, but this will be automatically corrected next time.", name))
			else
				self:AddQuest({
					id = id,
					complete = npcInfo,
				});
			end
		end
		
		i = i + 4;
	until (not name);
end

function core:QUEST_DETAIL()
	-- single quest offered directly, cache it
	local name = GetTitleText();
	
	if (name) then
		local npcInfo = self:UnitInfo("npc") or self:UnitInfo("questnpc") or {};
		local type = "npc";
		if (UnitGUID("npc") == UnitGUID("player")) then
			type = "trigger";
			npcInfo.npc = nil;
			npcInfo.npcid = nil;
		elseif (not npcInfo) then
			-- this is an item, we already logged it
			return
		end
		self:AddQuest({
			name = name,
			begin = npcInfo,
			type =  type,
		});
	end
end

function core:QUEST_COMPLETE()
	-- single quest offered directly, find the quest id in our log and add the end point 
	-- (note: assuming we are never going to have duplicate name quests in our log at one time)
	local name, id = GetTitleText(), self:GetQuestID(GetTitleText());
	
	if (not id) then
		-- self:Print(format("Cannot find quest id of '%s' from log", name))
		return
	end
	
	local npcInfo = self:UnitInfo("npc") or self:UnitInfo("questnpc") or {};
	self:AddQuest({
		id = id,
		complete = npcInfo,
	});	
end

function core:LOOT_OPENED()
	for i=1,GetNumLootItems() do
		local texture, item, quantity, quality, locked, isQuestItem, questId, isActive = GetLootSlotInfo(i);
		
		if (questId) then
			local npcInfo = self:UnitInfo("target");
			self:AddQuest({
				id = questId,
				begin = npcInfo,
				type = "item",
			});
		end
	end
end

function core:QUEST_ACCEPTED(_, index)
	local title, level, questTag, suggestedGroup, isHeader, isCollapsed, isComplete, isDaily, questID = GetQuestLogTitle(index);
	local db = self.db.global;
	
	if (not questID) then
		return
	end
		
	-- check name->id
	if (db.questsNoID[title]) then
		db.quests[questID] = db.questsNoID[title];
		db.questsNoID[title] = nil;
	end
	
	-- ammend information
	self:AddQuest({
		id = questID,
		name = title,
		level = level,
		cat = self:GetQuestZone(questID),
		daily = isDaily,
	});
end

function core:AddQuest(opt)
	local token = opt.id or opt.name or error("No token value specified");
	local db = tonumber(token) and self.db.global.quests or self.db.global.questsNoID;
	
	if (not db[token]) then
		db[token] = {};
		self.questCt = self.questCt + 1;
		self:CheckCount();
	end
	
	for i,v in pairs(opt) do
		db[token][i] = v;
	end
	
	-- Faction flag set
	db[token].faction = bit.bor( (db[token].faction or 0), FACTION_MAP[UnitFactionGroup("player")] );
end

function core:GetQuestZone(qid)
	local zone;
	
	for i=1,GetNumQuestLogEntries() do
		local title, _, _, _, isHeader, _, _, _, questID = GetQuestLogTitle(i);
		if (isHeader) then
			zone = title;
		elseif (qid == questID) then
			return zone or "Unknown";
		end	
	end
	
	return "Unknown";
end

function core:GetQuestID(name)
	for i=1,GetNumQuestLogEntries() do
		local title, _, _, _, _, _, _, _, questID = GetQuestLogTitle(i);
		
		if (title == name) then
			return questID;
		end
	end
	
	return nil;
end

-- UnitInfo(): returns table with {unit-name, object/npc-id, "obj" or "npc", player-zone, player-x-coord, player-y-coord}
function core:UnitInfo(unit)
	local guid = UnitGUID(unit);
	local type, id;
	if (guid) then
		if (guid:sub(5, 5) == "3") then
			-- npc
			type = "npc";
			id = tonumber(guid:sub(7, 10), 16);
		elseif (guid:sub(5, 5) == "1") then
			-- obj
			type = "obj";
			id = tonumber(guid:sub(6, 10), 16);
		end
	end		
	
	SetMapToCurrentZone();
	local x, y = GetPlayerMapPosition("player");
	return id and {
		npc = UnitName(unit), 
		npcid = id, 
		npctype = type, 
		zone = GetRealZoneText(), 
		zoneid = GetCurrentMapAreaID(),
		x = x,
		y = y,
	};
end

function core:CheckCount()
	if (self.questWarnThresh <= self.questCt) then
		self.questWarnThresh = self.questWarnThresh + 15000
		self:Print(format("You have logged %d quests; consider reloading your interface to flush the data. A fatal error in World of Warcraft will not save any data since your last reload.", self.questCt));
	end
end

function core:Debug(level, ...)
	if (type(level) ~= "number") then
		self:Debug(1, level, ...);
		return;
	end
	if self.db.profile.debug >= level then 
		self:Print(...);
	end
end



local wzid = {
   [1] = "DunMorogh",
   [3] = "Badlands",
   [4] = "BlastedLands",
   [8] = "SwampOfSorrows",
   [10] = "Duskwood",
   [11] = "Wetlands",
   [12] = "Elwynn",
   [14] = "Durotar",
   [15] = "Dustwallow",
   [16] = "Aszhara",
   [17] = "Barrens",
   [28] = "WesternPlaguelands",
   [33] = "Northern Stranglethorn",
   -- [36] = "Alterac Mountains",
   [38] = "LochModan",
   [40] = "Westfall",
   [41] = "DeadwindPass",
   [44] = "Redridge",
   [45] = "Arathi",
   [46] = "BurningSteppes",
   [47] = "Hinterlands",
   [51] = "SearingGorge",
   [65] = "Dragonblight",
   [66] = "ZulDrak",
   [67] = "TheStormPeaks",
   [85] = "Tirisfal",
   [130] = "Silverpine",
   [139] = "EasternPlaguelands",
   [141] = "Teldrassil",
   [148] = "Darkshore",
   [210] = "IcecrownGlacier",
   [215] = "Mulgore",
   [267] = "Hillsbrad Foothills",
   [331] = "Ashenvale",
   [357] = "Feralas",
   [361] = "Felwood",
   [394] = "GrizzlyHills",
   [400] = "ThousandNeedles",
   [405] = "Desolace",
   [406] = "StonetalonMountains",
   [440] = "Tanaris",
   [490] = "UngoroCrater",
   [493] = "Moonglade",
   [495] = "HowlingFjord",
   [616] = "Mount Hyjal",
   [618] = "Winterspring",
   [1377] = "Silithus",
   [1497] = "Undercity",
   [1519] = "Stormwind City",
   [1537] = "Ironforge",
   [1637] = "Orgrimmar",
   [1638] = "ThunderBluff",
   [1657] = "Darnassus",
   [3430] = "EversongWoods",
   [3433] = "Ghostlands",
   [3483] = "Hellfire",
   [3487] = "SilvermoonCity",
   [3518] = "Nagrand",
   [3519] = "TerokkarForest",
   [3520] = "ShadowmoonValley",
   [3521] = "Zangarmarsh",
   [3522] = "BladesEdgeMountains",
   [3523] = "Netherstorm",
   [3524] = "AzuremystIsle",
   [3525] = "BloodmystIsle",
   [3537] = "BoreanTundra",
   [3557] = "TheExodar",
   [3703] = "ShattrathCity",
   [3711] = "SholazarBasin",
   [4080] = "Sunwell",
   [4197] = "LakeWintergrasp",
   [4395] = "Dalaran",
   [4714] = "Gilneas",
   [4720] = "The Lost Isles",
   [4737] = "Kezan",
   [4755] = "Gilneas City",
   [4815] = "Kelp'thar Forest",
   [4922] = "Twilight Highlands",
   [5034] = "Uldum",
   [5042] = "Deepholm",
   [5095] = "Tol Barad",
   [5144] = "Shimmering Expanse",
   [5145] = "Abyssal Depths",
   [5146] = "Vashj'ir",
   [5416] = "The Maelstrom",
   [5630] = "The Maelstrom",
   [5695] = "Ahn'Qiraj: The Fallen Kingdom",
   -- [5706] = "TheSteamPools",
}


--[[local ltofile = {}
local l2tofile = {
	["Gilneas City"] = 611,
	["Gilneas"] = 545,
}
local c = {GetMapContinents()}
for cid in ipairs(c) do
   local z = {GetMapZones(cid)}
   for zid,zname in ipairs(z) do
      SetMapZoom(cid, zid)
      local file = GetMapInfo()
      ltofile[file] = GetCurrentMapAreaID()
	  l2tofile[zname] = GetCurrentMapAreaID()
   end
end

for k,v in pairs(wzid) do
   if ltofile[v] then
      wzid[k] = ltofile[v]
   elseif l2tofile[v] then
	  wzid[k] = l2tofile[v]
   else
      -- print("NO ZONE INFO FOR :", k, " - ", v)
      wzid[k] = nil
   end
end


local out = "local zidmap = {"
local sort = {}
for k,v in pairs(wzid) do
   table.insert(sort, k)
end
table.sort(sort)
for idx,num in ipairs(sort) do
   out = string.format("%s\n   %d: %d,", out, num, wzid[num])
end
out = out .. "\n}\n"

output = out]]